# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for post-processing of measurement results.

"""
import os

import pytest

from qbraid import device_wrapper
from qbraid.devices.result import _format_counts
from qbraid.interface import random_circuit

# Skip tests if IBM/AWS account auth/creds not configured
skip_remote_tests: bool = os.getenv("QBRAID_RUN_REMOTE_TESTS") is None
REASON = "QBRAID_RUN_REMOTE_TESTS not set (requires configuration of IBM/AWS storage)"


@pytest.mark.parametrize(
    "counts_raw, expected_out",
    [
        ({" 1": 474, "0": 550}, {"0": 550, "1": 474}),
        ({"10": 479, "1 1": 13, "0 0 ": 496}, {"00": 496, "01": 0, "10": 479, "11": 13}),
        (
            {"10 1": 586, "11  0  ": 139, "0 01": 496, "  010": 543, "11 1": 594},
            {
                "000": 0,
                "001": 496,
                "010": 543,
                "011": 0,
                "100": 0,
                "101": 586,
                "110": 139,
                "111": 594,
            },
        ),
    ],
)
def test_format_counts(counts_raw, expected_out):
    counts_out = _format_counts(counts_raw)
    assert counts_out == expected_out  # check equivalance
    assert list(counts_out.items()) == list(expected_out.items())  # check ordering of keys


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
@pytest.mark.parametrize("device_id", ["ibm_q_simulator_statevector", "aws_sv_sim"])
def test_result_wrapper_measurements(device_id):
    """Test result wrapper measurements method."""
    circuit = random_circuit("qiskit", num_qubits=3, depth=3, measure=True)
    sim = device_wrapper(device_id).run(circuit, shots=10)
    qbraid_result = sim.result()
    counts = qbraid_result.measurement_counts()
    measurements = qbraid_result.measurements()
    assert isinstance(counts, dict)
    assert measurements.shape == (10, 3)


@pytest.mark.skipif(skip_remote_tests, reason=REASON)
@pytest.mark.parametrize("device_id", ["ibm_q_qasm_simulator", "ibm_q_simulator_statevector"])
def test_result_wrapper_batch_measurements(device_id):
    """Test result wrapper measurements method for circuit batch."""
    circuit = random_circuit("qiskit", num_qubits=3, depth=3, measure=True)
    sim = device_wrapper(device_id).run_batch([circuit, circuit, circuit], shots=10)
    qbraid_result = sim.result()
    counts = qbraid_result.measurement_counts()
    measurements = qbraid_result.measurements()

    assert isinstance(counts, list)
    for count in counts:
        assert isinstance(count, dict)

    assert measurements.shape == (3, 10, 3)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
"""
Module defining Utility functions to be able to run IonQ device from AWS
"""
from typing import Union

import pytket
import pytket.extensions.braket
from braket.circuits import Circuit
from pytket._tket.circuit._library import _TK1_to_RzRx  # type: ignore
from pytket.passes import RebaseCustom
from pytket.predicates import (
    CompilationUnit,
    GateSetPredicate,
    MaxNQubitsPredicate,
    NoClassicalControlPredicate,
    NoFastFeedforwardPredicate,
    NoMidMeasurePredicate,
    NoSymbolsPredicate,
)

from ..wrappers import circuit_wrapper

HARMONY_MAX_QUBITS = 11

ionq_gates = {
    pytket.circuit.OpType.X,
    pytket.circuit.OpType.Y,
    pytket.circuit.OpType.Z,
    pytket.circuit.OpType.Rx,
    pytket.circuit.OpType.Ry,
    pytket.circuit.OpType.Rz,
    pytket.circuit.OpType.H,
    pytket.circuit.OpType.S,
    pytket.circuit.OpType.Sdg,
    pytket.circuit.OpType.T,
    pytket.circuit.OpType.Tdg,
    pytket.circuit.OpType.V,
    pytket.circuit.OpType.Vdg,
    pytket.circuit.OpType.Measure,
    pytket.circuit.OpType.noop,
    pytket.circuit.OpType.SWAP,
    pytket.circuit.OpType.CX,
    pytket.circuit.OpType.ZZPhase,
    pytket.circuit.OpType.XXPhase,
    pytket.circuit.OpType.YYPhase,
    pytket.circuit.OpType.ZZMax,
    pytket.circuit.OpType.Barrier,
}

preds = [
    NoClassicalControlPredicate(),
    NoFastFeedforwardPredicate(),
    NoMidMeasurePredicate(),
    NoSymbolsPredicate(),
    GateSetPredicate(ionq_gates),
    MaxNQubitsPredicate(HARMONY_MAX_QUBITS),
]

ionq_rebase_pass = RebaseCustom(
    ionq_gates,
    pytket.Circuit(),  # cx_replacement (irrelevant)
    _TK1_to_RzRx,
)  # tk1_replacement


def braket_ionq_compilation(circuit: Union[Circuit, pytket.circuit.Circuit]) -> Circuit:
    """
    Compiles a Braket circuit to a Braket circuit that can run on IonQ Harmony.

    Args:
        circuit (Union[braket.circuits.Circuit, pytket.circuit.Circuit]):
            The input Braket or PyTKET circuit to be compiled.

    Returns:
        braket.circuits.Circuit: The compiled Braket circuit that can run on IonQ Harmony.

    Notes:
        - If the input circuit is a braket Circuit, the function
          transpiles it to a ``pytket.circuit.Circuit`` before compilation.
        - The circuit is transpiled using qBraid's transpiler, if it contains
          any of the following gates::

                CPhaseShift00
                CPhaseShift01
                CPhaseShift10
                CV
                ECR
                GPi
                GPi2
                MS
                PSwap
                Unitary

        - Otherwise, the circuit is transpiled using ``pytket-braket``'s ``braket_to_tk``.

    """
    if isinstance(circuit, Circuit):
        try:
            tk_circuit = pytket.extensions.braket.braket_convert.braket_to_tk(circuit)
        except NotImplementedError:
            tk_circuit = circuit_wrapper(circuit).transpile("pytket")
    else:
        tk_circuit = circuit

    cu = CompilationUnit(tk_circuit, preds)
    ionq_rebase_pass.apply(cu)
    assert cu.check_all_predicates()
    compiled, _, _ = pytket.extensions.braket.braket_convert.tk_to_braket(cu.circuit)
    return compiled



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining IBMResultWrapper Class

"""
import numpy as np

from qbraid.devices.result import ResultWrapper


class IBMResultWrapper(ResultWrapper):
    """Qiskit ``Result`` wrapper class."""

    def _format_measurements(self, memory_list):
        """Format the measurements into int for the given memory list"""
        formatted_meas = []
        for str_shot in memory_list:
            lst_shot = [int(x) for x in list(str_shot)]
            formatted_meas.append(lst_shot)
        return formatted_meas

    def measurements(self):
        """Return measurements as list"""
        num_circuits = len(self.vendor_rlo.results)
        qiskit_meas = [self.vendor_rlo.get_memory(i) for i in range(num_circuits)]
        qbraid_meas = [self._format_measurements(qiskit_meas[i]) for i in range(num_circuits)]

        if num_circuits == 1:
            qbraid_meas = qbraid_meas[0]

        return np.array(qbraid_meas)

    def raw_counts(self):
        """Returns the histogram data of the run"""
        return self.vendor_rlo.get_counts()



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing qasm programs used for testing

"""
import os
from typing import Optional

import numpy as np

from qbraid.exceptions import QbraidError

QASMType = str
current_dir = os.path.dirname(os.path.abspath(__file__))


def qasm2_bell() -> QASMType:
    """Returns QASM2 bell circuit"""
    return open(os.path.join(current_dir, "bell.qasm"), mode="r", encoding="utf-8").read()


def qasm2_shared15():
    """Returns QASM2 for qBraid `TestSharedGates`."""
    return open(os.path.join(current_dir, "shared_15.qasm"), mode="r", encoding="utf-8").read()


def create_gateset_qasm(max_operands) -> np.ndarray:
    """gets qasm for gateset with max_operands"""

    q1_gates = [
        ("id", 1, 0),
        ("x", 1, 0),
        ("y", 1, 0),
        ("z", 1, 0),
        ("h", 1, 0),
        ("s", 1, 0),
        ("t", 1, 0),
        ("sdg", 1, 0),
        ("tdg", 1, 0),
        ("sx", 1, 0),
        ("rx", 1, 1),
        ("ry", 1, 1),
        ("rz", 1, 1),
        ("p", 1, 1),
        ("u1", 1, 1),
        ("u2", 1, 2),
        ("u3", 1, 3),
        ("reset", 1, 0),
    ]

    q2_gates = [
        ("cx", 2, 0),
        ("cy", 2, 0),
        ("cz", 2, 0),
        ("ch", 2, 0),
        ("cp", 2, 1),
        ("crx", 2, 1),
        ("cry", 2, 1),
        ("crz", 2, 1),
        ("swap", 2, 0),
        ("cu", 2, 4),
    ]

    q3_gates = [("ccx", 3, 0), ("cswap", 3, 0)]

    gates = q1_gates.copy()

    if max_operands >= 2:
        gates.extend(q2_gates)
    if max_operands >= 3:
        gates.extend(q3_gates)
    gates = np.array(
        gates, dtype=[("gate", object), ("num_qubits", np.int64), ("num_params", np.int64)]
    )
    return gates


def _qasm3_random(
    num_qubits: Optional[int] = None,
    depth: Optional[int] = None,
    max_operands: Optional[int] = None,
    seed=None,
    measure=False,
) -> QASMType:
    """Generate random QASM3 circuit string.

    Args:
        num_qubits (int): number of quantum wires
        depth (int): layers of operations (i.e. critical path length)
        max_operands (int): maximum size of gate for each operation
        seed (int): seed for random number generator
        measure (bool): whether to include measurement gates

    Raises:
        QbraidError: When invalid  random circuit options given

    Returns:
        QASM3 random circuit string

    """

    num_qubits = np.random.randint(1, 4) if num_qubits is None else num_qubits
    depth = np.random.randint(1, 4) if depth is None else depth
    max_operands = np.random.randint(1, 3) if max_operands is None else max_operands
    try:
        if seed is None:
            seed = np.random.randint(0, np.iinfo(np.int32).max)
        np.random.seed(seed)
        rng = np.random.default_rng(seed)
        # create random circuit qasm3.0
        qasm_code_header = f"""
// Random Circuit generated by qBraid
OPENQASM 3.0;
include "stdgates.inc";
/*
    seed = {seed}
    num_qubits = {num_qubits}
    depth = {depth}
    max_operands = {max_operands}
*/
"""
        max_operands = min(max_operands, num_qubits)
        if num_qubits == 0:
            rand_circuit = qasm_code_header
            return rand_circuit
        rand_circuit = qasm_code_header + f"qubit[{num_qubits}] q;\n"
        if measure:
            rand_circuit += f"bit[{num_qubits}] c;\n"
        qubits = np.arange(num_qubits)
        gates = create_gateset_qasm(max_operands)
        for _ in range(depth):
            gate_specs = rng.choice(gates, size=num_qubits)
            cumulative_qubits = np.cumsum(gate_specs["num_qubits"], dtype=np.int64)

            max_index = np.searchsorted(cumulative_qubits, num_qubits, side="right")
            gate_specs = gate_specs[:max_index]
            slack = num_qubits - cumulative_qubits[max_index - 1]
            if slack:
                gates = create_gateset_qasm(max_operands=1)
                slack_gates = rng.choice(gates, size=slack)
                gate_specs = np.hstack((gate_specs, slack_gates))

            q_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
            p_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
            q_indices[0] = p_indices[0] = 0
            np.cumsum(gate_specs["num_qubits"], out=q_indices[1:])
            np.cumsum(gate_specs["num_params"], out=p_indices[1:])
            parameters = rng.uniform(0, 2 * np.pi, size=p_indices[-1])
            for i, (gate, _, p) in enumerate(gate_specs):
                if p:
                    params = ",".join(
                        str(parameters[j]) for j in range(p_indices[i], p_indices[i + 1])
                    )
                    qubit_indices = ",".join(
                        f"q[{qubits[j]}]" for j in range(q_indices[i], q_indices[i + 1])
                    )
                    line = f"{gate}({params}) {qubit_indices};\n"
                    rand_circuit += line
                else:
                    qubit_indices = ",".join(
                        f"q[{qubits[j]}]" for j in range(q_indices[i], q_indices[i + 1])
                    )
                    line = f"{gate} {qubit_indices};\n"
                    rand_circuit += line
            qubits = rng.permutation(qubits)
        if measure:
            for i in range(num_qubits):
                rand_circuit += f"c[{i}] = measure q[{i}];\n"

    except Exception as e:
        raise QbraidError("Could not create Qasm random circuit") from e
    return rand_circuit



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QiskitCircuitWrapper Class

"""
from cirq import Circuit

from qbraid.transpiler.wrappers.abc_qprogram import QuantumProgramWrapper


class QiskitCircuitWrapper(QuantumProgramWrapper):
    """Wrapper class for Qiskit ``Circuit`` objects"""

    def __init__(self, circuit: Circuit):
        """Create a QiskitCircuitWrapper

        Args
            circuit: the qiskit ``Circuit`` object to be wrapped
        """
        super().__init__(circuit)

        self._qubits = circuit.qubits
        self._params = circuit.parameters
        self._num_qubits = circuit.num_qubits
        self._num_clbits = circuit.num_clbits
        self._depth = circuit.depth()
        self._package = "qiskit"
        self._program_type = "QuantumCircuit"



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining input / output types for a quantum backend:

  * QPROGRAM: Type alias defining all supported quantum circuit / program types

  * QPROGRAM_LIBS: List of all supported quantum software libraries / packages

"""
from typing import Union

import braket.circuits
import cirq
import pyquil
import pytket.circuit
import qiskit

# Supported quantum programs.
QASMType = str

QPROGRAM = Union[
    braket.circuits.Circuit,
    cirq.Circuit,
    qiskit.QuantumCircuit,
    pyquil.Program,
    pytket.circuit.Circuit,
    QASMType,
]

_PROGRAMS = [
    braket.circuits.Circuit,
    cirq.Circuit,
    qiskit.QuantumCircuit,
    pyquil.Program,
    pytket.circuit.Circuit,
]


# pylint: disable-next=bad-str-strip-call
_PROGRAM_TYPES = [str(x).strip("<class").strip(">").strip(" ").strip("'") for x in _PROGRAMS]
QPROGRAM_TYPES = _PROGRAMS + [QASMType]

_PROGRAM_LIBS = [x.split(".")[0] for x in _PROGRAM_TYPES]
QPROGRAM_LIBS = _PROGRAM_LIBS + ["qasm2"]



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for interacting with the qBraid Jobs API.

"""
import os
import sys
from datetime import datetime
from typing import TYPE_CHECKING

from .session import QbraidSession

if TYPE_CHECKING:
    import qbraid

SLUG = "qbraid_sdk_9j9sjy"  # qBraid Lab environment ID.
ENVS_PATH = os.getenv("QBRAID_USR_ENVS") or os.path.join(
    os.path.expanduser("~"), ".qbraid", "environments"
)
SLUG_PATH = os.path.join(ENVS_PATH, SLUG)


def _running_in_lab():
    """Checks if you are running qBraid-SDK in qBraid Lab environment.

    See https://docs.qbraid.com/en/latest/lab/environments.html
    """
    python_exe = os.path.join(SLUG_PATH, "pyenv", "bin", "python")
    return sys.executable == python_exe


def _qbraid_jobs_enabled():
    """Returns True if running qBraid Lab and qBraid Quantum Jobs
    proxy is enabled. Otherwise, returns False.

    See https://docs.qbraid.com/en/latest/lab/quantumjobs.html
    """
    proxy_file = os.path.join(SLUG_PATH, "qbraid", "proxy")
    if os.path.isfile(proxy_file):
        with open(proxy_file) as f:  # pylint: disable=unspecified-encoding
            firstline = f.readline().rstrip()
            return "active = true" in firstline  # check if proxy is active or not
    return False


def init_job(
    vendor_job_id: str,
    device: "qbraid.devices.DeviceLikeWrapper",
    circuits: "qbraid.transpiler.QuantumProgramWrapper",
    shots: int,
) -> str:
    """Initialize data dictionary for new qbraid job and
    create associated MongoDB job document.

    Args:
        vendor_job_id: Job ID provided by device vendor
        device: Wrapped quantum device
        circuit: Wrapped quantum circuit list
        shots: Number of shots

    Returns:
        The qbraid job ID associated with this job

    """
    session = QbraidSession()

    # One of the features of qBraid Quantum Jobs is the ability to send
    # jobs without any credentials using the qBraid Lab platform. If the
    # qBraid Quantum Jobs proxy is enabled, a document has already been
    # created for this job. So, instead creating a duplicate, we query the
    # user jobs for the `vendorJobId` and return the correspondong `qbraidJobId`.
    if _running_in_lab() and _qbraid_jobs_enabled():
        job = session.post("/get-user-jobs", json={"vendorJobId": vendor_job_id}).json()[0]
        return job["qbraidJobId"]

    # Create a new document for the user job. The qBraid API creates a unique
    # Job ID, which is collected in the response. We use dummy variables for
    # each of the status fields, which will be updated via the `get_job_data`
    # function upon instantiation of the `JobLikeWrapper` object.
    init_data = {
        "qbraidJobId": "",
        "vendorJobId": vendor_job_id,
        "qbraidDeviceId": device.id,
        "vendorDeviceId": device.vendor_device_id,
        "shots": shots,
        "createdAt": datetime.utcnow(),
        "status": "UNKNOWN",  # this will be set after we get back the job ID and check status
        "qbraidStatus": "INITIALIZING",  # TODO use qbraid Enums for status values
        "email": os.getenv("JUPYTERHUB_USER") or session.user_email,
    }

    if len(circuits) == 1:
        init_data["circuitNumQubits"] = circuits[0].num_qubits
        init_data["circuitDepth"]: circuits[0].depth
    else:
        init_data["circuitBatchNumQubits"] = ([circuit.num_qubits for circuit in circuits],)
        init_data["circuitBatchDepth"] = [circuit.depth for circuit in circuits]

    return session.post("/init-job", data=init_data).json()


def get_job_data(qbraid_job_id: str, update: dict = None) -> dict:
    """Update a new MongoDB job document.

    Args:
        qbraid_job_id: The qbraid job ID associated with the job
        status: Job status update

    Returns:
        The metadata associated with this job

    """
    session = QbraidSession()
    body = {"qbraidJobId": qbraid_job_id}
    # Two status variables so we can track both qBraid and vendor status.
    if update is not None and "status" in update and "qbraidStatus" in update:
        body["status"] = update["status"]
        body["qbraidStatus"] = update["qbraidStatus"]
    metadata = session.put("/update-job", data=body).json()[0]
    metadata.pop("_id", None)
    metadata.pop("user", None)
    return metadata



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQasm tools

"""
import os
import re

import numpy as np
from cirq.circuits import Circuit
from qiskit.circuit import QuantumCircuit

from qbraid.transpiler.cirq_qasm.qasm_conversions import from_qasm, to_qasm
from qbraid.transpiler.cirq_qasm.qelib1_defs import _decompose_rxx_instr

QASMType = str


def qasm_qubits(qasmstr: str) -> QASMType:
    """get number of qasm qubits"""

    return [
        text.replace("\n", "")
        for match in re.findall(r"(\bqreg\s\S+\s+\b)|(qubit\[(\d+)\])", qasmstr)
        for text in match
        if text != "" and len(text) >= 2
    ]


def qasm_num_qubits(qasmstr: str) -> QASMType:
    """calculate number of qubits"""
    q_num = 0

    for num in qasm_qubits(qasmstr):
        q_num += int(re.search(r"\d+", num).group())
    return q_num


def qasm_depth(qasmstr: str) -> QASMType:
    """calculate number of depth"""
    circuit = from_qasm(qasmstr)
    return len(Circuit(circuit.all_operations()))


def _convert_to_contiguous_qasm(qasmstr: str, rev_qubits=False) -> QASMType:
    """delete qubit with no gate and optional reverse circuit"""
    # pylint: disable=import-outside-toplevel
    from qbraid.interface.qbraid_cirq.tools import _convert_to_contiguous_cirq

    circuit = to_qasm(_convert_to_contiguous_cirq(from_qasm(qasmstr), rev_qubits=rev_qubits))
    return circuit


def _unitary_from_qasm(qasmstr: QASMType) -> np.ndarray:
    """Return the unitary of the QASM"""
    return from_qasm(qasmstr).unitary()


def _build_qasm_3_reg(line: str, qreg_type: bool) -> QASMType:
    """Helper function to build openqasm 3 register statements

    Args:
        line (str): openqasm 2  regdecl statement
        qreg_type (bool): whether a qreg or creg type statement

    Returns:
        str : openqasm 3 qubits / bits declaration
    """
    reg_keyword_len = 4
    line = line[reg_keyword_len:]
    elements = line.split("[")
    reg_name = elements[0].strip()
    reg_size = elements[1].split("]")[0].strip()
    result = "qubit" if qreg_type else "bit"
    result += f"[{reg_size}] {reg_name};\n"
    return result


def _build_qasm_3_measure(line: str) -> QASMType:
    """Helper function to build openqasm 3 measure string

    Args:
        line (str): openqasm 2 measure statement

    Returns:
        str:  openqasm 3 measure statement
    """
    measure_keyword_len = 7
    line = line[measure_keyword_len:]
    elements = line.split("->")
    qubits_name = elements[0].replace(" ", "")
    bits_name = elements[1].split(";")[0].replace(" ", "")

    return f"{bits_name} = measure {qubits_name};\n"


def _change_to_qasm_3(line: str) -> QASMType:
    """Function to change an openqasm 2 line to openqasm 3

    Args:
        line (str): an openqasm 2 line

    Returns:
        str: corresponding openqasm 3 line
    """
    line = line.lstrip()
    if line.startswith("OPENQASM"):
        return ""
    if "qelib1.inc" in line:
        return ""
    if line.startswith("qreg"):
        return _build_qasm_3_reg(line, qreg_type=True)
    if line.startswith("creg"):
        return _build_qasm_3_reg(line, qreg_type=False)
    if line.startswith("u("):
        return line.replace("u(", "U(")
    if line.startswith("rxx("):
        return _decompose_rxx_instr(line)
    if line.startswith("measure"):
        return _build_qasm_3_measure(line)
    if line.startswith("opaque"):
        # as opaque is ignored by openqasm 3 add it as a comment
        return "// " + line + "\n"
    return line + "\n"


def convert_to_qasm3(qasm_2_str: str):
    """Convert a QASM 2.0 string to QASM 3.0 string

    Args:
        qasm_2_str (str): QASM 2.0 string
    """
    try:
        # use inbuilt method to check validity
        _ = QuantumCircuit.from_qasm_str(qasm_2_str)
    except Exception as e:
        raise ValueError("Invalid QASM 2.0 string") from e

    #  a newline separated qasm 2 string
    # formatted_qasm_2 = circuit.qasm()
    qasm_3_str = """OPENQASM 3.0;
include 'stdgates.inc';"""

    # add the gate from qelib1.inc not present in the
    # stdgates.inc file
    current_dir = os.path.dirname(os.path.abspath(__file__))
    with open(
        os.path.join(current_dir, "qasm_lib/qelib_qasm3.qasm"), mode="r", encoding="utf-8"
    ) as gate_defs:
        for line in gate_defs:
            qasm_3_str += line

    for line in qasm_2_str.splitlines():
        line = _change_to_qasm_3(line)
        qasm_3_str += line
    return qasm_3_str



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

import logging
import os

import pytest
from qiskit.circuit import QuantumCircuit
from qiskit.qasm3 import loads

from qbraid.interface import circuits_allclose, random_circuit
from qbraid.interface.qbraid_qasm.circuits import _qasm3_random, qasm2_bell, qasm2_shared15
from qbraid.interface.qbraid_qasm.tools import (
    convert_to_qasm3,
    qasm_depth,
    qasm_num_qubits,
    qasm_qubits,
)


def test_qasm_qubits():
    """test calculate qasm qubit"""

    assert qasm_qubits(qasm2_bell()) == ["qreg q[2];"]
    assert qasm_qubits(qasm2_shared15()) == ["qreg q[4];"]


def test_qasm_num_qubits():
    assert qasm_num_qubits(qasm2_bell()) == 2
    assert qasm_num_qubits(qasm2_shared15()) == 4


def test_qasm_depth():
    """test calcualte qasm depth"""
    assert qasm_depth(qasm2_bell()) == 2
    assert qasm_depth(qasm2_shared15()) == 22


def _check_output(output, expected):
    actual_circuit = loads(output)
    expected_circuit = loads(expected)
    assert actual_circuit == expected_circuit


@pytest.mark.parametrize(
    "num_qubits, depth, max_operands, seed, measure",
    [
        (
            None,
            None,
            None,
            None,
            False,
        ),  # Test case 1: Generate random circuit with default parameters
        (3, 3, 3, 42, False),  # Test case 2: Generate random circuit with specified parameters
        (None, None, None, None, True),  # Test case 3: Generate random circuit with measurement
    ],
)
def test_qasm3_random(num_qubits, depth, max_operands, seed, measure):
    """Test random circuit generation using _qasm_random"""

    circuit = _qasm3_random(
        num_qubits=num_qubits, depth=depth, max_operands=max_operands, seed=seed, measure=measure
    )

    assert qasm_num_qubits(circuit) >= 1
    if num_qubits is not None:
        assert qasm_num_qubits(circuit) == num_qubits


def test_qasm3_random_with_known_seed():
    # Test case 4: Generate a random circuit with measurement with known Seed and compare with expected output
    circuit = _qasm3_random(num_qubits=3, depth=3, max_operands=3, seed=42, measure=True)
    assert qasm_num_qubits(circuit) == 3

    out__expected = """
// Random Circuit generated by qBraid
OPENQASM 3.0;
include "stdgates.inc";
/*
    seed = 42
    num_qubits = 3
    depth = 3
    max_operands = 3
*/
qubit[3] q;
bit[3] c;
y q[0];
crx(5.3947298351621535) q[1],q[2];
cz q[0],q[2];
t q[1];
cp(0.8049616944763924) q[1],q[2];
u1(2.829858307545725) q[0];
c[0] = measure q[0];
c[1] = measure q[1];
c[2] = measure q[2];

"""
    _check_output(circuit, out__expected)


def test_convert_to_qasm3():
    """test the conversion of qasm 2 to 3"""
    lib_dir = os.path.dirname(os.path.dirname(__file__)) + "/qasm_lib"
    gate_def_qasm3 = open(
        os.path.join(lib_dir, "qelib_qasm3.qasm"), mode="r", encoding="utf-8"
    ).read()

    # 1. qubit statement conversion
    test_qregs = """OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[1] ;
    qreg qubits  [10]   ;
    creg c[1];
    creg bits   [12]   ;
    """
    test_qregs_expected = f"""OPENQASM 3.0;
    include "stdgates.inc";
    {gate_def_qasm3}
    qubit[1] q;
    qubit[10] qubits;
    bit[1] c;
    bit[12] bits;"""

    _check_output(convert_to_qasm3(test_qregs), test_qregs_expected)

    # 2. Measure statement conversion
    test_measure = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    creg c[2];
    measure q->c;
    measure q[0] -> c[1];
    """

    test_measure_expected = f"""
    OPENQASM 3.0;
    include "stdgates.inc";
    {gate_def_qasm3}
    qubit[2] q;
    bit[2] c;
    c = measure q;
    c[1] = measure q[0];
    """
    _check_output(convert_to_qasm3(test_measure), test_measure_expected)

    # 3. Opaque comment conversion
    test_opaque = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    creg c[2];
    opaque custom_gate (a,b,c) p,q,r;
    """

    test_opaque_expected = f"""
    OPENQASM 3.0;
    include "stdgates.inc";
    {gate_def_qasm3}
    qubit[2] q;
    bit[2] c;
    // opaque custom_gate (a,b,c) p,q,r;
    """

    _check_output(convert_to_qasm3(test_opaque), test_opaque_expected)

    # 4. std header change
    test_header = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[1];
    """

    test_header_expected = f"""
    OPENQASM 3.0;
    include "stdgates.inc";
    {gate_def_qasm3}
    qubit[1] q;
    """
    _check_output(convert_to_qasm3(test_header), test_header_expected)

    # 5. Unsupported gate conversion
    test_unsupported = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[5];
    u(1,2,3) q[0];
    sxdg q[0];
    csx q[0], q[1];
    cu1(0.5) q[0], q[1];
    cu3(1,2,3) q[0], q[1];
    rzz(0.5) q[0], q[1];
    rccx q[0], q[1], q[2];
    rc3x q[0], q[1], q[2], q[3];
    c3x q[0], q[1], q[2], q[3];
    c3sqrtx q[0], q[1], q[2], q[3];
    c4x q[0], q[1], q[2], q[3], q[4];
    """
    test_unsupported_expected = f"""
    OPENQASM 3.0;   
    include "stdgates.inc";
    {gate_def_qasm3}
    qubit[5] q;
    U(1,2,3) q[0];
    sxdg q[0];
    csx q[0], q[1];
    cu1(0.5) q[0], q[1];
    cu3(1,2,3) q[0], q[1];
    rzz(0.5) q[0], q[1];
    rccx q[0], q[1], q[2];
    rc3x q[0], q[1], q[2], q[3];
    c3x q[0], q[1], q[2], q[3];
    c3sqrtx q[0], q[1], q[2], q[3];
    c4x q[0], q[1], q[2], q[3], q[4];
    """
    _check_output(convert_to_qasm3(test_unsupported), test_unsupported_expected)


def _generate_valid_qasm_strings(seed=42, gates_to_skip=None, num_circuits=100):
    """Returns a list of 100 random qasm2 strings
    which do not contain any of the gates in gates_to_skip

    Current list of invalid gates is ["u", "cu1", "cu2", "cu3", "rxx"]
    For the motivation, see discussion
                   - https://github.com/Qiskit/qiskit-qasm3-import/issues/12
                   - https://github.com/Qiskit/qiskit-qasm3-import/issues/11#issuecomment-1568505732
    """
    if gates_to_skip is None:
        gates_to_skip = []

    qasm_strings = []
    while len(qasm_strings) < num_circuits:
        try:
            circuit_random = random_circuit("qiskit", seed=seed)
            qasm_str = circuit_random.qasm()
            circuit_from_qasm = QuantumCircuit.from_qasm_str(qasm_str)
        except Exception as e:
            logging.error("Invalid QASM generated by random_circuit: %s", e)
            continue

        for gate in gates_to_skip:
            if len(circuit_from_qasm.get_instructions(gate)) > 0:
                break
        else:
            qasm_strings.append(qasm_str)

    return qasm_strings


@pytest.mark.parametrize("qasm2_str", _generate_valid_qasm_strings(gates_to_skip=["r"]))
def test_random_conversion_to_qasm3(qasm2_str):
    """test random gates conversion"""
    qasm3_str = convert_to_qasm3(qasm2_str)
    circuit_orig = QuantumCircuit.from_qasm_str(qasm2_str)
    circuit_test = loads(qasm3_str)

    # ensure that the conversion is correct
    assert circuits_allclose(circuit_orig, circuit_test)


@pytest.mark.skip(reason="Qiskit terra bug")
def test_u0_gate_conversion():
    """test u0 gate conversion
    Separate test due to bug in terra ,
    see https://github.com/Qiskit/qiskit-terra/issues/10184
    """

    lib_dir = os.path.dirname(os.path.dirname(__file__)) + "/qasm_lib"
    gate_def_qasm3 = open(
        os.path.join(lib_dir, "qelib_qasm3.qasm"), mode="r", encoding="utf-8"
    ).read()

    test_u0 = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[1];
    u0(0.5) q[0];"""

    test_u0_expected = f"""
    OPENQASM 3.0;
    include "stdgates.inc";
    {gate_def_qasm3}
    qubit[1] q;
    u0(0.5) q[0];
    """

    _check_output(convert_to_qasm3(test_u0), test_u0_expected)


def test_rxx_gate_conversion():
    """test rxx gate conversion
    Separate test due to bug in qasm3 lib,
    see https://github.com/Qiskit/qiskit-qasm3-import/issues/11

    Once resolved, add definition to -
        ``qbraid/interface/qbraid_qasm/qasm_lib/lib/qelib_qasm3.qasm``
    """

    lib_dir = os.path.dirname(os.path.dirname(__file__)) + "/qasm_lib"
    gate_def_qasm3 = open(
        os.path.join(lib_dir, "qelib_qasm3.qasm"), mode="r", encoding="utf-8"
    ).read()

    test_rxx = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[3];
    rxx(0.5) q[0], q[1];"""

    test_rxx_expected = f"""
    OPENQASM 3.0;
    include "stdgates.inc";
    {gate_def_qasm3}
    qubit[3] q;

    // rxx gate
    h q[0];
    h q[1];
    cx q[0],q[1];
    rz(0.5) q[1];
    cx q[0],q[1];
    h q[1];
    h q[0];
    """

    _check_output(convert_to_qasm3(test_rxx), test_rxx_expected)



# Copyright (C) 2023 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.
import string

import braket
import numpy as np
import pytest
import scipy

from qbraid.devices.ionq import braket_ionq_compilation

#############
### UTILS ###
#############


def generate_params(varnames):
    params = {}
    for v in varnames:
        if v.startswith("angle"):
            params[v] = np.random.rand() * 2 * np.pi
    return params


def get_braket_gates():
    braket_gates = {
        attr: None for attr in dir(braket.circuits.Gate) if attr[0] in string.ascii_uppercase
    }
    for gate in ["C", "PulseGate"]:
        braket_gates.pop(gate)

    for gate in braket_gates:
        if gate == "Unitary":
            n = np.random.randint(1, 4)
            unitary = scipy.stats.unitary_group.rvs(2**n)
            braket_gates[gate] = getattr(braket.circuits.Gate, gate)(matrix=unitary)
        else:
            params = generate_params(
                getattr(braket.circuits.Gate, gate).__init__.__code__.co_varnames
            )
            braket_gates[gate] = getattr(braket.circuits.Gate, gate)(**params)

    return {k: v for k, v in braket_gates.items() if v is not None}


#############
### TESTS ###
#############

braket_gates = get_braket_gates()


@pytest.mark.parametrize("gate_name", braket_gates)
def test_braket_ionq_compilation(gate_name):
    gate = braket_gates[gate_name]
    if gate.qubit_count == 1:
        source_circuit = braket.circuits.Circuit([braket.circuits.Instruction(gate, 0)])
    else:
        source_circuit = braket.circuits.Circuit(
            [braket.circuits.Instruction(gate, range(gate.qubit_count))]
        )

        braket_ionq_compilation(
            source_circuit
        )  # the function already has an assertion which checks that the predicates are satistfied
